using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using FarseerPhysics;
using FarseerPhysics.Common;
using FarseerPhysics.DebugViews;
using FarseerPhysics.Dynamics;
using FarseerPhysics.SamplesFramework;
using Flatland.AgentBehavior;
using Flatland.Elements;
using Flatland.FarseerSamples;
using Flatland.FarseerSamples.ScreenSystem;
using Flatland.GeneticProgramming;
using Flatland.GeneticProgramming.Compilation;
using Flatland.GeneticProgramming.CoreTypes;
using Flatland.GeneticProgramming.FileWriter;
using Flatland.GeneticProgramming.TestProblems_Do_not_modify_;
using Flatland.Geomerty;
using Flatland.Input;
using Flatland.Labs;
using Flatland.Utilities;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Color = Microsoft.Xna.Framework.Color;

/**
* 
* Copyright (c) 2012 Anders Hst Kjrgaard | ahkj@itu.dk
* 
* Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
* associated documentation files (the "Software"), to deal in the Software without restriction, 
* including without limitation the rights to use, copy, modify, merge, publish, distribute, 
* sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all copies or substantial 
* portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT 
* NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

namespace Flatland {
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class FlatlandWorld : Game {
        /// <summary>
        /// Default scaling with camera zoom.
        /// </summary>
        public const float CAMERA_ZOOM_SCALE = 0.05f;

        /// <summary>
        /// World size from the centre
        /// </summary>
        public const float WorldSize = 40f;

        /// <summary>
        /// In Flatland there is a small constant attraction to the south
        /// </summary>
        public const float FLATLAND_SOUTHWARD_ATTRACTION = 0.00042f;

        /// <summary>
        /// Default Camera pan speed.
        /// </summary>
        public const float CAMERA_PAN_SPEED = 0.4f;

        // public const int MAX_GAME_SPEED = 9;

        /// <summary>
        /// Default population size.
        /// </summary>
        public const int DEFAULT_POPULATION_SIZE = 100;

        /// <summary>
        /// Max population
        /// </summary>
        public const int MAX_POPULATION_SIE = 1000;

        /// <summary>
        /// Default time between reproduction in seconds.
        /// </summary>
        public const float DEFAULT_REPRODUCTIONS_TIME = 10f;

        /// <summary>
        /// Default expected life time in seconds.
        /// </summary>
        public const int DEFAULT_EXPECTED_LIFE_TIME = 120;

        public const int WORLD_FOOD = 200;

        /// <summary>
        /// Private. Population size.
        /// </summary>
        uint _populationSize;

        /// <summary>
        /// GP for reuse of strategies.
        /// </summary>
        GP _gp;

        /// <summary>
        /// DateTime of the start if the simulation
        /// </summary>
        private DateTime _runstart = DateTime.Now;

        /// <summary>
        /// Control panel.
        /// </summary>
        public PanelController ControlPanel;

        //XNA
        public GraphicsDeviceManager GraphicsDeviceManager;

        public static int TotalAgentsSpawned = 0;

        /// <summary>
        /// Reference to the FPE World
        /// </summary>
        public World World;

        /// <summary>
        /// Assets from Farseer samples. Methods for creating textures with different kinds of geometrical shapes.
        /// </summary>
        public AssetCreator Assets;

        /// <summary>
        /// Spritebatch. Used for most drawing relative to the Camera.
        /// </summary>
        public SpriteBatch SpriteBatch;


        /// <summary>
        /// The reproduction time.
        /// </summary>
        /// <value>
        /// The reproduction time.
        /// </value>
        public float ReproductionTime = DEFAULT_REPRODUCTIONS_TIME;

        /// <summary>
        /// Count till next reproduction.
        /// </summary>
        float _reproductionTimeCount = 1;


        /// <summary>
        /// Get the linebatch. We just reuse this spritebatch everywhere.
        /// </summary>
        public LineBatch LineBatch;

        /// <summary>
        /// Camera field.
        /// </summary>
        public Camera2D Camera;

        /// <summary>
        /// Flatlab instance can be set for running lab experiments.
        /// </summary>
        public FlatlandLab FlatlandLab;

        /// <summary>
        /// All food placed in the world
        /// </summary>
        public List<Food> FoodReserve;

        /// <summary>
        /// The amount of food in Flatland.
        /// </summary>
        public int FoodAmount = 15;

        /// <summary>
        /// The background color on Flatland
        /// </summary>
        Color BackGroundColor = Color.CornflowerBlue;

        /// <summary>
        /// The male population of flatlanders at a given time.
        /// </summary>
        public ArrayList Flatlanders = new ArrayList(DEFAULT_POPULATION_SIZE);


        /// <summary>
        /// Private. List of all female flatlanders
        /// </summary>
        ArrayList _flatlanderines;

        /// <summary>
        /// The woman population of flatlanders at a given time.
        /// </summary>
        public ArrayList Flatlanderines {
            get { return _flatlanderines ?? (_flatlanderines = new ArrayList((int)_populationSize)); }
            set { _flatlanderines = value; }
        }


        // Farseer expects objects to be scaled to MKS (meters, kilos, seconds)
        // 1 meters equals 64 pixels here
        // (Objects should be scaled to be between 0.1 and 10 meters in size)
        //#source http://farseerphysics.codeplex.com/discussions/275582
        private const float MeterInPixels = 64f;

        private Vector2 _screenCentre;
        private DebugViewXNA _debugView;

        /// <summary>
        /// Construct Flatland. Pass a FlatLab instance to simulate lab.
        /// </summary>
        /// <param name="lab"></param>
        public FlatlandWorld(FlatlandLab lab = null) {
            if (lab != null) {
                FlatlandLab = lab;
                FlatlandLab.Flatland = this;
            }
            _gp = new GP(new CommandTestGene());
            _populationSize = _gp.PopulationSize;
            Init();
        }

        private Stopwatch updateSW, drawSW;

        /// <summary>
        /// Init things that needs to be done before Initialize during construction.
        /// </summary>
        void Init() {
            Content.RootDirectory = "Content";
            Window.Title = "Flatland";
            var d = Directory.CreateDirectory(Program.MY_GENES_FOLDER);
            d.Attributes &= ~FileAttributes.ReadOnly;
            GraphicsDeviceManager = new GraphicsDeviceManager(this);
            GraphicsDeviceManager.PreferredBackBufferWidth = 950;
            GraphicsDeviceManager.PreferredBackBufferHeight = 700;
            GraphicsDeviceManager.IsFullScreen = false;
            GraphicsDeviceManager.ApplyChanges();
            ControlPanel = new PanelController();
            //  Console.WriteLine("0.0 {0} \n Width: {1}\n height {2}", ConvertUnits.ToSimUnits(new Vector2(0, 0))
            //      , ConvertUnits.ToSimUnits(new Vector2(0, 0)), ConvertUnits.ToSimUnits(new Vector2(GraphicsDeviceManager.PreferredBackBufferWidth, GraphicsDeviceManager.PreferredBackBufferHeight)));
            IsFixedTimeStep = false;
            IsMouseVisible = true;
            drawSW = new Stopwatch();
            updateSW = new Stopwatch();
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize() {
            FoodReserve = new List<Food>(FoodAmount);
            Camera = new Camera2D(GraphicsDevice);

            World = new World(new Vector2(0f, FLATLAND_SOUTHWARD_ATTRACTION));
            SpriteBatch = new SpriteBatch(GraphicsDevice);
            LineBatch = new LineBatch(GraphicsDevice);
            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent() {
            Assets = new AssetCreator(this);
            Assets.LoadContent(Content);
            ControlPanel.LoadContent();
            if (FlatlandLab != null) {
                FlatlandLab.LoadContent();
            }
            else {
                SetupFlatLand();

            }
            Camera.Zoom -= CAMERA_ZOOM_SCALE * 5;
            Camera.Position = Vector2.Zero;

            // create and configure the debug view
            _debugView = new DebugViewXNA(World);
            _debugView.AppendFlags(DebugViewFlags.DebugPanel);
            _debugView.DefaultShapeColor = Color.White;
            _debugView.SleepingShapeColor = Color.LightGray;
            _debugView.LoadContent(GraphicsDevice, Content);
            _screenCentre = new Vector2(GraphicsDevice.Viewport.Width / 2f,
                                                GraphicsDevice.Viewport.Height / 2f);
            base.LoadContent();
        }

        private int i = 0;
        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// tick time 00:00:00.0166667
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime) {
            InputHandler.HandleMouse();
            InputHandler.HandleKeyboard();
            ControlPanel.Update();
            Camera.Update(gameTime);


            if (ReproductionTime < _reproductionTimeCount) {
                BreedAgent();
                _reproductionTimeCount = 0;
            }
            _reproductionTimeCount += (float)gameTime.ElapsedGameTime.TotalSeconds;

            if (FlatlandLab != null)
                FlatlandLab.Update(gameTime);

            for (int j = 0; j < Flatlanders.Count; j++) {
                ((Agent)Flatlanders[j]).Update();
            }

            for (int j = 0; j < Flatlanderines.Count; j++) {
                ((Agent)Flatlanderines[j]).Update();
            }

            if (FoodReserve.Count == 0 && FlatlandLab == null) {
                ReplenishAndKill();
            }
            base.Update(gameTime);
            World.Step(Math.Min((float)gameTime.ElapsedGameTime.TotalSeconds, (1f / 30f)));
        }

        /// <summary>
        /// Food is up so flatlanders with no food eaten starve to death.
        /// New food is the replenished.
        /// </summary>
        public void ReplenishAndKill() {
            for (int j = 0; j < Flatlanders.Count; j++) {
                if (((Agent)Flatlanders[j]).FoodEaten == 0)
                    KillAgent(((Agent)Flatlanders[j]));
                else
                    --((Agent)Flatlanders[j]).FoodEaten;
            }

            for (int j = 0; j < Flatlanderines.Count; j++) {
                if (((Agent)Flatlanderines[j]).FoodEaten == 0)
                    KillAgent(((Agent)Flatlanderines[j]));
                else
                    --((Agent)Flatlanderines[j]).FoodEaten;
            }
            FoodReserve = Food.SetupFood(this, false, WORLD_FOOD, WorldSize);
        }

        /// <summary>
        /// Show info on a followed Flatlander. Removes panel if null.
        /// </summary>
        /// <param name="body"></param>
        public void Follow(Body body) {
            Camera.TrackingBody = body;
            ControlPanel.Show((Agent)body.UserData);
        }

        /// <summary>
        /// Unfollows any agent being followed and hides the control panel.
        /// </summary>
        public void Unfollow() {
            Camera.TrackingBody = null;
            ControlPanel.Hide();
        }

        /// <summary>
        /// Randomly set a new background color
        /// </summary>
        public void ChangeBackgroundColor() {
            var colors = Utils.SystemColors();
            BackGroundColor = colors[Utils.Random.Next(colors.Length)];
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime) {
            //   drawSW.Start();
            GraphicsDevice.Clear(BackGroundColor);
            SpriteBatch.Begin(SpriteSortMode.BackToFront,
                         BlendState.AlphaBlend,
                         null,
                         null,
                         null,
                         null,
                         Camera.View);

            foreach (Food food in FoodReserve) {
                food.Draw();
            }
            foreach (Agent flatlander in Flatlanders) {
                flatlander.Draw();
            }
            foreach (Agent flatlanderine in Flatlanderines) {
                flatlanderine.Draw();
            }
            if (FlatlandLab != null)
                FlatlandLab.Draw();

            SpriteBatch.End();
            ControlPanel.Draw();
            // DrawDebugView();
            base.Draw(gameTime);
        }

        void DrawDebugView() {
            // calculate the projection and view adjustments for the debug view
            Matrix projection = Matrix.CreateOrthographicOffCenter(0f, GraphicsDevice.Viewport.Width / MeterInPixels,
                                                             GraphicsDevice.Viewport.Height / MeterInPixels, 0f, 0f,
                                                             1f);

            Matrix view = Matrix.CreateTranslation(new Vector3((Camera.Position / MeterInPixels) - (_screenCentre / MeterInPixels), 0f)) * Matrix.CreateTranslation(new Vector3((_screenCentre / MeterInPixels), 0f));
            // draw the debug view
            _debugView.RenderDebugData(ref projection, ref view);
        }

        public void RemoveFood() {
            foreach (Food food in FoodReserve) {
                World.RemoveBody(food.FoodBody);
            }
            FoodReserve.RemoveAll(x => x != null);
        }

        /// <summary>
        /// Prints an Agent. Possibly from the control panel.
        /// </summary>
        public void PrintAgent() {
            if (!ControlPanel.Hiding && ControlPanel.Agent != null) {
                string fileName = "\\Flatlander_" + ControlPanel.Agent.Id;
                string outPutFolder = "Flatland_" + _runstart.ToFileTime() + fileName;
                GeneFileWriter gfw = new GeneFileWriter();
                var d = Directory.CreateDirectory(outPutFolder);
                d.Attributes &= ~FileAttributes.ReadOnly;
                Compiler comp = new Compiler(outPutFolder);
                foreach (var report in ControlPanel.Agent.Reports) {
                    gfw.Serialize(report, outPutFolder);
                    //The compiler recompiles the gneotype, but we do not want to destroy the one we have, so we 
                    //hand it a clone
                    comp.Compile((Report)report.Clone(), report.GenoType.GetType().Name, true);
                }
                int count = 0;
                foreach (var entry in Directory.GetFileSystemEntries(outPutFolder)) {
                    if (entry.Contains("\\Flatlander_"))
                        count++;
                }
                if (count != 0)
                    fileName += "_" + count;
                ControlPanel.news_output = outPutFolder;
                Stream stream = new FileStream(outPutFolder + fileName + ".png", FileMode.Create, FileAccess.Write | FileAccess.Read, FileShare.None);
                ControlPanel.Agent.Texture.SaveAsPng(stream, ControlPanel.Agent.Texture.Width, ControlPanel.Agent.Texture.Height);
                stream.Close();
            }
        }

        /// <summary>
        /// Spreads an amount of food over Flatland.
        /// </summary>
        /// <param name="amount"></param>
        public void Feed(int amount = 100) {
            FoodReserve.AddRange(Food.SetupFood(this, false, amount, 20));
        }

        /// <summary>
        /// Inheritance rule for a Flatlander
        /// TODO add inhertance rule for Isosceles fathers
        /// </summary>
        /// <param name="agent"></param>
        /// <returns></returns>
        public int InheritVerts(Agent father) {
            return father.Vertices() + 1;
        }

        /// <summary>
        /// Breeds an Agent and inserts it in Flatland. 
        /// Requires that at least one man and one woman is alive
        /// </summary>
        /// <param name="genes">Fixed set of genes.</param>
        public void BreedAgent(List<Report> genes = null) {
            Agent newAgent = new Agent();
            //Verts must be between 3 and max verts
            int verts = GetInitialVerts(3);
            if (genes == null) {
                if (Flatlanderines.Count == 0) return;
                if (Flatlanders.Count == 0) return;
                var mother = (Agent)Flatlanderines[Utils.Random.Next(Flatlanderines.Count)];
                var father = (Agent)Flatlanders[Utils.Random.Next(Flatlanders.Count)];
                genes = BehaviorFactory.EvolveBehavior(father, mother, _gp);
                //Its a boy!
                verts = InheritVerts(father);
            }
            if (Utils.Random.Next(3) == 0)
                verts = 2;//Its a girl!

            newAgent.Id = TotalAgentsSpawned + "";
            if (Flatlanders.Count == 0 && Flatlanderines.Count == 0)
                newAgent.Control = true;
            ++TotalAgentsSpawned;

            Behavior.BehaviorFactory(newAgent, genes);
            float x = ConvertUnits.ToSimUnits(Utils.Random.Next((int)ConvertUnits.ToDisplayUnits(WorldSize) * 2) - (int)ConvertUnits.ToDisplayUnits(WorldSize));
            float y = ConvertUnits.ToSimUnits(Utils.Random.Next((int)ConvertUnits.ToDisplayUnits(WorldSize) * 2) - (int)ConvertUnits.ToDisplayUnits(WorldSize));
            newAgent.Size = Utils.Random.Next(Agent.MAX_SIZE + Agent.MIN_SIZE) - Agent.MIN_SIZE;
            newAgent.InitAgent(new Vector2(x, y), verts);
            if (TotalAgentsSpawned == 0 && UserPreferences.VERBOSE)
                newAgent.Control = true;
            if (verts == 2)
                Flatlanderines.Add(newAgent);
            else
                Flatlanders.Add(newAgent);
        }

        /// <summary>
        /// Initial amount of verts.
        /// </summary>
        /// <returns></returns>
        private int GetInitialVerts(int minVerts) {
            if (Utils.Random.Next(5) == 0 || minVerts == Agent.MAX_VERTICES) return minVerts;
            return GetInitialVerts(minVerts + 1);
        }

        /// <summary>
        /// Show the info panel.
        /// </summary>
        public void ShowInfoPanel() {
            ControlPanel.Hide();
            ControlPanel.ShowInfo();

        }

        /// <summary>
        /// Setup Flatland with a number of flantlanders and things
        /// </summary>
        public void SetupFlatLand() {
            if (Flatlanders.Count != 0)
                return;
            for (int i = 0; i < DEFAULT_POPULATION_SIZE; i++) {
                BreedAgent(BehaviorFactory.CreateBaseBehavior());
            }
            FoodReserve = Food.SetupFood(this, false, WORLD_FOOD, WorldSize);
        }

        /// <summary>
        /// Kill an Agent.
        /// </summary>
        /// <param name="agent"></param>
        public void KillAgent(Agent agent) {
            if (agent.IsWoman)
                Flatlanderines.Remove(agent);
            else
                Flatlanders.Remove(agent);
            agent.RemoveAgent();
        }
    }
}
