﻿using System;
using System.Collections.Generic;
using Flatland.GeneticProgramming.Compilation;
using Flatland.GeneticProgramming.CoreTypes;
using Flatland.GeneticProgramming.Crossovers;
using Flatland.GeneticProgramming.Initializations;
using Flatland.GeneticProgramming.SyntaxTree;
using Flatland.GeneticProgramming.TestProblems_Do_not_modify;
using Flatland.GeneticProgramming.TestProblems_Do_not_modify_;
using Flatland.Utilities;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace Test {
    [TestClass]
    public class CrossOverTest : TestHelper {
        /// <summary>
        /// Test crossover with a mix of primitive types. 
        /// No nodes must exist in two trees.
        /// Children must have a parent that knows them.
        /// Primitives must be with correct nodes - primitives in terminals, and functions in nodes.
        /// Trees are never deeper than max depth
        /// TODO clean parameters to crossover
        /// </summary>
        [TestMethod]
        public void CrossoverMixedExpressionTrees() {
            foreach (InitializationTechnique initializationTechnique in Enum.GetValues(typeof(InitializationTechnique))) {
                Initialization init = new Initialization(initializationTechnique);
                foreach (CrossoverTechnique crossoverTechnique in Enum.GetValues(typeof(CrossoverTechnique))) {
                    int popSize = 5; //will be times 3
                    int maxDepth = 5;
                    int gens = 50;
                    List<Report> pop = new List<Report>(popSize * 3);

                    //GenoType 1
                    GenoType gene = new Flatland.GeneticProgramming.TestProblems_Do_not_modify.Test();
                    PrimitiveSet primitiveSet = new PrimitiveSet();
                    primitiveSet.Parse(gene);
                    for (int i = 0; i < popSize; i++) {
                        Report r = new Report();
                        r.GenoType = (GenoType)gene.Clone();
                        r.GenoType.Tree = init.Init((uint) maxDepth, primitiveSet);
                        pop.Add(r);
                    }

                    //GenoType 2 with different function names
                    GenoType gene2 = new TestGene3();
                    PrimitiveSet prim2 = new PrimitiveSet();
                    prim2.Parse(gene2);
                    for (int i = 0; i < popSize; i++) {
                        Report r = new Report();
                        r.GenoType = (GenoType)gene.Clone();
                        r.GenoType.Tree = init.Init((uint)maxDepth, prim2);
                        pop.Add(r);
                    }

                    //Genotype 3 with similar functions as prop 1
                    GenoType gene3 = new Flatland.GeneticProgramming.TestProblems_Do_not_modify.Test();
                    PrimitiveSet DNA3 = new PrimitiveSet();
                    DNA3.Parse(gene3);
                    Tree t3 = init.Init(3, DNA3);

                    ICrossOver crosser = new Crossover(crossoverTechnique);
                    for (int i = 0; i < gens; i++) {
                        //perform crossover
                        crosser.Cross(pop, 1, 4);
                        CheckPopulationForUniqueNodes(pop);
                        foreach (Report report in pop) {
                            Assert.IsTrue(report.GenoType.Tree.MaxDetpth() <= maxDepth);
                            CheckAllHasParent(report.GenoType.Tree);
                            CheckCorrectPrimitivesInNodes(report.GenoType.Tree);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Crossover of command trees. 
        /// No nodes must exist in two trees.
        /// Children must have a parent that knows them.
        /// Primitives must be with correc nodes. Primitives in terminals, and functions in nodes.
        /// Trees are never deeper than max depth
        /// </summary>
        [TestMethod]
        public void CrossOverCommandTrees() {
            foreach (InitializationTechnique initializationTechnique in Enum.GetValues(typeof(InitializationTechnique))) {
                Initialization init = new Initialization(initializationTechnique);
                foreach (CrossoverTechnique crossoverTechnique in Enum.GetValues(typeof(CrossoverTechnique))) {
                    for(int maxdepth = 1; maxdepth <= 4; maxdepth++)
                    {
                        int gens = 50;
                        int popSize = 5;

                        List<Report> pop = new List<Report>(popSize);

                        //CommandType
                        GenoType prop = new TestGeneCommand();
                        PrimitiveSet primitiveSet = new PrimitiveSet();
                        primitiveSet.Parse(prop);
                        //Create pop
                        for (int i = 0; i < popSize; i++) {
                            Report r = new Report();
                            r.GenoType = (GenoType)prop.Clone();
                            r.GenoType.Tree = init.Init((uint) maxdepth, primitiveSet);
                            pop.Add(r);
                        }

                        ICrossOver crosser = new Crossover(crossoverTechnique);
                        for (int i = 0; i < gens; i++) {

                            //perform crossover
                            crosser.Cross(pop, 1, (int)maxdepth);
                            CheckPopulationForUniqueNodes(pop);
                            foreach (Report report in pop) {
                                Assert.IsTrue(report.GenoType.Tree.MaxDetpth() <= maxdepth);
                                CheckAllHasParent(report.GenoType.Tree);
                                CheckCorrectPrimitivesInNodes(report.GenoType.Tree);
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Test runs with all cross over strategies. All strategies are invoked with probability 1.
        /// Crossover with probability 1 must produce new valid trees and cross at least one node if the 
        /// the available functions sets allows for crossover.
        /// Several tree depths are testedd
        /// Bloat control is not checked.
        /// </summary>
        /// 
        [TestMethod]
        public void TestCrossover() {
            foreach (InitializationTechnique initializationTechnique in Enum.GetValues(typeof(InitializationTechnique)))
            {
                Initialization init = new Initialization(initializationTechnique);
                Compiler comp = new Compiler("Crossover_Output");
                foreach (CrossoverTechnique crossoverTechnique in Enum.GetValues(typeof (CrossoverTechnique)))
                {
                    uint treeDepth = 3;
                    //GenoType 1
                    GenoType prop = new Flatland.GeneticProgramming.TestProblems_Do_not_modify.Test();
                    PrimitiveSet primitiveSet = new PrimitiveSet();
                    primitiveSet.Parse(prop);
                    Tree t1 = init.Init(treeDepth, primitiveSet);

                    //GenoType 2 with different arity in methods
                    GenoType prop2 = new TestGene3();
                    PrimitiveSet dna2 = new PrimitiveSet();
                    dna2.Parse(prop2);
                    Tree t2 = init.Init(treeDepth, dna2);
                    
                    //Add to report
                    List<Report> reports = new List<Report>(2);
                    Report a = new Report();
                    a.GenoType = prop;
                    a.GenoType.Tree = t1;
                    reports.Add(a);

                    Report b = new Report();
                    b.GenoType = prop2;
                    b.GenoType.Tree = t2;
                    reports.Add(b);

                    ICrossOver crosser = new Crossover(crossoverTechnique);
                    Console.WriteLine("Crossover: {0} Initialization: {1}", crossoverTechnique, initializationTechnique);
                    Console.WriteLine("Using example genes {0} and {1}", prop, prop2);
                    for (int i = 0; i < 50; i++)
                    {
                        Tree t1Before = (Tree) t1.Clone();
                        Tree t2Before = (Tree) t2.Clone();
                        Console.WriteLine("BEFORE tree cross over:");
                        Console.WriteLine("Tree 1:");
                        reports[0].GenoType.Tree.PrintTree();
                        Console.WriteLine("Tree 2:");
                        reports[1].GenoType.Tree.PrintTree();
                        crosser.Cross(reports, 1, (int) treeDepth);
                        Console.WriteLine("AFTER Tree cross over:");
                        reports[0].GenoType.Tree.PrintTree();
                        Console.WriteLine("Tree 2:");
                        reports[1].GenoType.Tree.PrintTree();
                        Console.WriteLine();
                        Console.WriteLine();
                        Console.WriteLine("**************************");
                        Console.WriteLine();

                        Assert.IsFalse(CheckHasSharedReference(t1, t2), "The same node was found in both trees after crossover");
                        CheckAllHasParent(t1);
                        CheckAllHasParent(t2);
                        CheckCorrectPrimitivesInNodes(t1);
                        CheckCorrectPrimitivesInNodes(t2);
                        if(i == 49)
                        {
                            Console.WriteLine("T1 Final max depth " + t1.MaxDetpth());
                            Console.WriteLine("T2 Final max depth " + t2.MaxDetpth());
                        }
                    }
                }
            }
        }
    }
}
