﻿using System;
using Flatland.GeneticProgramming;
using Flatland.GeneticProgramming.Crossovers;
using Flatland.GeneticProgramming.Initializations;
using Flatland.GeneticProgramming.Mutations;
using Flatland.GeneticProgramming.Selections;
using Flatland.GeneticProgramming.TestProblems_Do_not_modify_;
using Microsoft.VisualStudio.TestTools.UnitTesting;

/**
* 
* Copyright (c) 2012 Anders Høst Kjærgaard | ahkj@itu.dk
* 
* Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
* associated documentation files (the "Software"), to deal in the Software without restriction, 
* including without limitation the rights to use, copy, modify, merge, publish, distribute, 
* sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all copies or substantial 
* portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT 
* NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

namespace Test {
    public class TestAgent {
        public string Name { get; set; }
        public int Age { get; set; }
    }

    /// <summary>
    /// Test suite for GP. Looks at datastructures and corrct runs of the GP. 
    /// </summary>
    [TestClass]
    public class GpRunTest : TestHelper {

        /// <summary>
        /// Test run runs the GP sobject to all GPMethods in the system. Assertions are made for each iteration.
        /// Mutation and crossover rate is set to .5
        /// Running this test takes time.
        /// 
        /// For each evolment it must hold that:
        /// No report, gene or tree exists twice in the population.
        /// No gene carries a tree node that exists in another tree
        /// No tree is deeper than the max depth
        /// All trees can be executed
        /// All trees have a valid structure         
        /// </summary>
        [TestMethod]
        public void TestRunExpr() {
            const double fitnessgoal = int.MaxValue;
            const uint generations = 100;
            foreach (InitializationTechnique initializationTechnique in Enum.GetValues(typeof(InitializationTechnique))) {
                Initialization init = new Initialization(initializationTechnique);
                foreach (CrossoverTechnique crossoverTechnique in Enum.GetValues(typeof(CrossoverTechnique))) {
                    Crossover crossover = new Crossover(crossoverTechnique);
                    foreach (MutationTechnique mutationTechnique in Enum.GetValues(typeof(MutationTechnique))) {
                        Mutation mutation = new Mutation(mutationTechnique);

                        foreach (SelectionTechnique selectionmethod in Enum.GetValues(typeof(SelectionTechnique))) {
                            Flatland.GeneticProgramming.TestProblems_Do_not_modify.Test test = new Flatland.GeneticProgramming.TestProblems_Do_not_modify.Test();
                            var run = new GP(test, fitnessgoal, 50, 5, generations, mutationTechnique, 0.5f,
                                                selectionmethod, crossoverTechnique, 0.5f, initializationTechnique);
                            run.IgnoresOutput = true;
                            run.EvaluateFitness = false;
                            run.Init();
                            while (run.Evolve()) {
                                for (int i = 0; i < run.Population.Count; i++) {
                                    var individual = run.Population[i];
                                    CheckAllHasParent(individual.GenoType.Tree);
                                    CheckCorrectPrimitivesInNodes(individual.GenoType.Tree);
                                    Assert.IsTrue(individual.GenoType.Tree.MaxDetpth() <= run.MaxTreeDepth);
                                    individual.GenoType.Execute();
                                    for (int j = 0; j < run.Population.Count; j++) {
                                        if (i != j) {
                                            Assert.IsTrue(run.Population[j] != individual);
                                            Assert.IsTrue(run.Population[j].GenoType != individual.GenoType);
                                            Assert.IsTrue(run.Population[j].GenoType.Tree != individual.GenoType.Tree);
                                        }
                                    }
                                }
                                CheckPopulationForUniqueNodes(run.Population);
                            }
                            Console.WriteLine("Testrun " + initializationTechnique + "," + crossover + "," + mutationTechnique + "," + selectionmethod);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Test run runs the GP sobject to all GPMethods in the system. Assertions are made for each iteration.
        /// Mutation and crossover rate is set to .5
        /// Running this test takes time.
        /// 
        /// For each evolment it must hold that:
        /// No report, gene or tree exists twice in the population.
        /// No gene carries a tree node that exists in another tree
        /// No tree is deeper than the max depth
        /// All trees can be executed
        /// All trees have a valid structure      
        /// </summary>
        [TestMethod]
        public void TestRunCmd() {
            const double fitnessgoal = int.MaxValue;
            const uint generations = 100;
            foreach (InitializationTechnique initializationTechnique in Enum.GetValues(typeof(InitializationTechnique))) {
                Initialization init = new Initialization(initializationTechnique);
                foreach (CrossoverTechnique crossoverTechnique in Enum.GetValues(typeof(CrossoverTechnique))) {
                    Crossover crossover = new Crossover(crossoverTechnique);
                    foreach (MutationTechnique mutationTechnique in Enum.GetValues(typeof(MutationTechnique))) {
                        Mutation mutation = new Mutation(mutationTechnique);

                        foreach (SelectionTechnique selectionTechnique in Enum.GetValues(typeof(SelectionTechnique))) {
                            CommandTestGene testGene = new CommandTestGene();
                            var run = new GP(testGene, fitnessgoal, 50, 5, generations, mutationTechnique, 0.5f,
                                                selectionTechnique, crossoverTechnique, 0.5f, initializationTechnique);
                            run.IgnoresOutput = true;
                            run.EvaluateFitness = false;
                            run.Init();
                            while (run.Evolve()) {
                                for (int i = 0; i < run.Population.Count; i++) {
                                    var individual = run.Population[i];
                                    CheckAllHasParent(individual.GenoType.Tree);
                                    CheckCorrectPrimitivesInNodes(individual.GenoType.Tree);
                                    Assert.IsTrue(individual.GenoType.Tree.MaxDetpth() <= run.MaxTreeDepth);
                                    individual.GenoType.Execute();
                                    for (int j = 0; j < run.Population.Count; j++) {
                                        if (i != j) {
                                            Assert.IsTrue(run.Population[j] != individual);
                                            Assert.IsTrue(run.Population[j].GenoType != individual.GenoType);
                                            Assert.IsTrue(run.Population[j].GenoType.Tree != individual.GenoType.Tree);
                                        }
                                    }
                                }
                                CheckPopulationForUniqueNodes(run.Population);
                            }
                            Console.WriteLine("Testrun " + initializationTechnique + "," + crossover + "," + mutationTechnique + "," + selectionTechnique + "ok");
                        }
                    }
                }
            }
        }


        /// <summary>
        /// Can be used to check on run results.
        /// </summary>
        [TestMethod]
        public void FastTestRun() {
            int t = 42 * 42 * 42 * 42;
            uint gen = 5;
            uint pop = 5;
            var cerebro = new GP(new Flatland.GeneticProgramming.TestProblems_Do_not_modify.Test(), t, pop, 3, gen);
            cerebro.Run();
            Console.WriteLine("Fast Gp run done");
        }
    }
}
