﻿using System.Collections.Generic;
using Flatland.GeneticProgramming.CoreTypes;
using Flatland.GeneticProgramming.SyntaxTree;
using Microsoft.VisualStudio.TestTools.UnitTesting;

/**
* 
* Copyright (c) 2012 Anders Høst Kjærgaard | ahkj@itu.dk
* 
* 
* Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
* associated documentation files (the "Software"), to deal in the Software without restriction, 
* including without limitation the rights to use, copy, modify, merge, publish, distribute, 
* sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all copies or substantial 
* portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT 
* NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

namespace Test {
    /// <summary>
    /// Base class for most tests.
    /// For now mainly helper methods for checking that trees are valid
    /// </summary>
    [TestClass]
    public class TestHelper {

        /// <summary>
        /// True if two trees are logically equalivant. 
        /// Every node must have the same amount of children, child nodes must be at the same index and true by Equals
        /// </summary>
        protected bool CheckLogicalEqualityInNodes(Tree t1, Tree t2) {
            bool b = true;
            if (t1.GetType() == typeof(Node) && t2.GetType() == typeof(Node)) {
                if (((Node)t1).Children.Count == ((Node)t2).Children.Count) {
                    for (int i = 0; i < ((Node)t1).Children.Count; i++) {
                        if (!CheckLogicalEqualityInNodes(((Node)t1).Children[i], ((Node)t2).Children[i])) {
                            b = false;
                            break;
                        }
                    }
                }
                else b = false; //Unequal amount of children
            }
            return b && t1.Equals(t2);
        }

        /// <summary>
        /// True if a node exists in both tree.
        /// </summary>
        /// <param name="t1"></param>
        /// <param name="t2"></param>
        /// <returns></returns>
        protected bool CheckHasSharedReference(Tree t1, Tree t2) {
            if (t1.GetType() == typeof(Leaf)) return CheckNodeExists(t1, t2);
            bool b = false;
            foreach (var child in ((Node)t1).Children) {
                if (CheckHasSharedReference(child, t2)) {
                    b = true;
                    break;
                }
            }
            return b || CheckNodeExists(t1, t2);
        }

        /// <summary>
        /// Returns true if a node exists in a tree
        /// </summary>
        /// <param name="node">Node to be found</param>
        /// <param name="tree">Searched tree</param>
        /// <returns></returns>
        protected bool CheckNodeExists(Tree node, Tree tree) {
            if (tree.GetType() == typeof(Leaf)) return node == tree;
            bool b = false;
            foreach (var child in ((Node)tree).Children) {
                if (CheckNodeExists(node, child)) {
                    b = true;
                    break;
                }
            }
            return b || node == tree;
        }

        /// <summary>
        /// Check that every child node has a reference to its actual parent node.
        /// </summary>
        /// <param name="tree"></param>
        protected void CheckAllHasParent(Tree tree) {
            //Check arg is not a root node
            if (tree.Depth() != 1) {
                _CheckAllHasParent(tree, tree.Parent);
            }
            else if (tree.GetType() == typeof(Node)) {
                foreach (var child in ((Node)tree).Children) {
                    _CheckAllHasParent(child, tree);
                }
            }
        }

        /// <summary>
        /// Check that every non root-node has a parent that is not null and that the parent
        /// has the node as its child.
        /// </summary>
        /// <param name="tree"></param>
        /// <param name="parent"></param>
        private void _CheckAllHasParent(Tree tree, Tree parent) {
            Assert.IsNotNull(tree.Parent, "Parent node was null");
            Assert.IsTrue(tree.Parent == parent, "Node had wrong reference to its parent");
            Assert.IsTrue(((Node)tree.Parent).Children.Exists(t => t == tree), "Parent did not know its child");
            if (tree.GetType() == typeof(Node)) {
                foreach (Tree child in ((Node)tree).Children) {
                    _CheckAllHasParent(child, tree);
                }
            }
        }

        /// <summary>
        /// Asserts that all Node types carry a Function and Leaf carry a Terminal
        /// </summary>
        /// <param name="tree"></param>
        protected void CheckCorrectPrimitivesInNodes(Tree tree) {
            Assert.IsNotNull(tree, "Tree node was null");
            if (tree.GetType() == typeof(Leaf))
                Assert.IsTrue(((Leaf)tree).Terminal.GetType() == typeof(Terminal), "Leaf did not contain a terminal");
            if (tree.GetType() == typeof(Node)) {
                Assert.IsTrue(((Node)tree).Function.GetType() == typeof(Function), "Node did not contain a function");
                foreach (Tree child in ((Node)tree).Children) {
                    CheckCorrectPrimitivesInNodes(child);
                }
            }
        }

        /// <summary>
        /// Checks every node in every tree against every node in every other tree. Fails if an equal reference is found.
        /// </summary>
        /// <param name="pop"></param>
        protected void CheckPopulationForUniqueNodes(List<Report> pop) {
            //Check all node references.
            for (int j = 0; j < pop.Count; j++) {
                for (int k = 0; k < pop.Count; k++) {
                    if (j != k)
                        Assert.IsFalse(CheckHasSharedReference(pop[j].GenoType.Tree, pop[k].GenoType.Tree));
                }
            }
        }
    }
}
